@@ -1,7 +1,9 @@
 package Complete::Path;
 
+our $AUTHORITY = 'cpan:PERLANCAR'; # AUTHORITY
 our $DATE = '2021-02-02'; # DATE
-our $VERSION = '0.250'; # VERSION
+our $DIST = 'Complete-Path'; # DIST
+our $VERSION = '0.251'; # VERSION
 
 use 5.010001;
 use strict;
@@ -120,12 +122,21 @@ _
         #},
         recurse => {
             schema => 'bool*',
-            description => <<'_',
-
-Note that in recurse mode, leaf digging is not done.
-
-_
+            cmdline_aliases => {r=>{}},
         },
+        recurse_matching => {
+            schema => ['str*', in=>['level-by-level', 'all-at-once']],
+            default => 'level-by-level',
+        },
+        exclude_leaf => {
+            schema => 'bool*',
+        },
+        exclude_dir => {
+            schema => 'bool*',
+        },
+    },
+    args_rels => {
+        dep_all => [recurse_matching => ['recurse']],
     },
     result_naked => 1,
     result => {
@@ -144,6 +155,9 @@ sub complete_path {
     my $result_prefix = $args{result_prefix};
     my $starting_path = $args{starting_path} // '';
     my $recurse = $args{recurse};
+    my $recurse_matching = $args{recurse_matching} // 'level-by-level';
+    my $exclude_leaf = $args{exclude_leaf};
+    my $exclude_dir  = $args{exclude_dir};
 
     my $ci          = $Complete::Common::OPT_CI;
     my $word_mode   = $Complete::Common::OPT_WORD_MODE;
@@ -154,6 +168,67 @@ sub complete_path {
 
     my $re_ends_with_path_sep = qr!\A\z|\Q$path_sep\E\z!;
 
+    my @res;
+
+    my $cut_chars;
+    if (defined $args{_cut_chars}) {
+        $cut_chars = $args{_cut_chars};
+    } else {
+        $cut_chars = 0;
+        if (length($starting_path)) {
+            $cut_chars += length($starting_path);
+            unless ($starting_path =~ /\Q$path_sep\E\z/) {
+                $cut_chars += length($path_sep);
+            }
+        }
+    }
+
+  RECURSE_MATCHING_ALL_AT_ONCE: {
+        # recurse matching all-at-once is way simpler, we just need to collect
+        # all the nodes, then complate against it.
+        last unless $recurse && $recurse_matching eq 'all-at-once';
+        my @dirs = ($starting_path);
+        while (@dirs) {
+            my $dir = shift @dirs;
+            my $listres = $list_func->($dir, '', 0);
+            next unless $listres && @$listres;
+          L1:
+            for my $e (@$listres) {
+                my $p = $dir =~ $re_ends_with_path_sep ?
+                    "$dir$e" : "$dir$path_sep$e";
+
+                {
+                    local $_ = $p; # convenience for filter func
+                    next L1 if $filter_func && !$filter_func->($p);
+                }
+
+                my $is_dir;
+                if ($e =~ $re_ends_with_path_sep) {
+                    $is_dir = 1;
+                } else {
+                    local $_ = $p; # convenience for is_dir_func
+                    $is_dir = $is_dir_func->($p);
+                }
+
+                if ($is_dir) { push @dirs, $p }
+
+                # format result
+                $p = "$result_prefix$p" if length($result_prefix);
+                substr($p, 0, $cut_chars) = '' if $cut_chars;
+                unless ($p =~ /\Q$path_sep\E\z/) {
+                    $p .= $path_sep if $is_dir;
+                }
+
+                push @res, $p unless ($is_dir && $exclude_dir) || (!$is_dir && $exclude_leaf);
+            } # entry
+        } # while dirs
+        @res = @{ Complete::Util::complete_array_elem(
+            array => \@res,
+            word  => $word,
+        ) };
+        goto RETURN_RESULT;
+    }
+
     # split word by into path elements, as we want to dig level by level (needed
     # when doing case-insensitive search on a case-sensitive tree).
     my @intermediate_dirs;
@@ -236,20 +311,6 @@ sub complete_path {
     }
     log_trace "[comppath] candidate paths: %s", \@candidate_paths if $ENV{COMPLETE_PATH_TRACE};
 
-    my $cut_chars;
-    if (defined $args{_cut_chars}) {
-        $cut_chars = $args{_cut_chars};
-    } else {
-        $cut_chars = 0;
-        if (length($starting_path)) {
-            $cut_chars += length($starting_path);
-            unless ($starting_path =~ /\Q$path_sep\E\z/) {
-                $cut_chars += length($path_sep);
-            }
-        }
-    }
-
-    my @res;
     for my $dir (@candidate_paths) {
         #say "D:opendir($dir)";
         my $listres = $list_func->($dir, $leaf, 0);
@@ -311,10 +372,12 @@ sub complete_path {
             unless ($p =~ /\Q$path_sep\E\z/) {
                 $p .= $path_sep if $is_dir;
             }
-            push @res, $p, @subres;
+            push @res, $p unless ($is_dir && $exclude_dir) || (!$is_dir && $exclude_leaf);
+            push @res, @subres;
         }
     }
 
+  RETURN_RESULT:
     \@res;
 }
 1;
@@ -332,7 +395,7 @@ Complete::Path - Complete path
 
 =head1 VERSION
 
-This document describes version 0.250 of Complete::Path (from Perl distribution Complete-Path), released on 2021-02-02.
+This document describes version 0.251 of Complete::Path (from Perl distribution Complete-Path), released on 2021-02-02.
 
 =head1 DESCRIPTION
 
@@ -362,6 +425,10 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
+=item * B<exclude_dir> => I<bool>
+
+=item * B<exclude_leaf> => I<bool>
+
 =item * B<filter_func> => I<code>
 
 Provide extra filtering. Code will be given path and should return 1 if the item
@@ -395,7 +462,7 @@ separator by C<complete_path()>.
 
 =item * B<recurse> => I<bool>
 
-Note that in recurse mode, leaf digging is not done.
+=item * B<recurse_matching> => I<str> (default: "level-by-level")
 
 =item * B<starting_path>* => I<str> (default: "")
 
